#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "./ivec.h"

struct Ivec {
  int *data;       /* dynamic memory area containing the integers */
  size_t length;   /* how many integers are currently stored in data */
  size_t capacity; /* how many integers can be stored in data */
};

Ivec *ivec_new() {
  Ivec *xs = malloc(sizeof(Ivec));
  if (xs == NULL) {
    return NULL;
  }
  xs->capacity = 1;
  xs->length = 0;
  xs->data = malloc(xs->capacity * sizeof(int));
  if (xs->data == NULL) {
    free(xs);
    return NULL;
  }
  return xs;
}

void ivec_free(Ivec *xs) {
  free(xs->data);
  free(xs);
}

#define ivec_at_internal(xs, i) ((xs)->data+(i))

int *ivec_at(Ivec *xs, size_t i) { return ivec_at_internal(xs,i); }

size_t ivec_length(Ivec *xs) { return xs->length; }

size_t ivec_capacity(Ivec *xs) { return xs->capacity; }

bool ivec_set_capacity(Ivec *xs, size_t capacity) {
  if (xs->length <= capacity) {
    xs->capacity = capacity;
    xs->data = realloc(xs->data, capacity * sizeof(int));
    return xs->data != NULL;
  }
  return true;
}

bool ivec_push(Ivec *xs, int x) {
  if (xs->capacity == xs->length) {
    bool success = ivec_set_capacity(xs, xs->capacity * 2);
    if (!success) {
      return false;
    }
  }
  xs->data[xs->length] = x;
  xs->length++;
  return true;
};

void ivec_pop(Ivec *xs) {
  if (xs->length > 0) {
    xs->length--;
    if (xs->length * 2 == xs->capacity) {
      ivec_set_capacity(xs, xs->length);
    }
  }
}

void swap_int(int *x, int *y) {
  int tmp = *x;
  *x = *y;
  *y = tmp;
}

int *ivec_min_between(Ivec *xs, size_t from, size_t to) {
  if (from >= to || ivec_length(xs) < to) {
    return NULL;
  }
  int *min = ivec_at_internal(xs, from);
  for (size_t i = from + 1; i < to; ++i) {
    int *candidate = ivec_at_internal(xs, i);
    if (*candidate < *min) {
      min = candidate;
    }
  }
  return min;
}

void ivec_sort(Ivec *xs) {
  size_t len = ivec_length(xs);
  for (size_t i = 0; i < len; ++i) {
    swap_int(ivec_min_between(xs, i, len), ivec_at_internal(xs, i));
  }
}

void ivec_print(Ivec *xs) {
  printf("Vector at address %p has %ld elements and capacity %ld.\n", xs,
         ivec_length(xs), ivec_capacity(xs));
  for (size_t i = 0; i < ivec_length(xs); ++i) {
    int *xi = ivec_at_internal(xs, i);
    printf("  ivec[%ld] = %d (address %p)\n", i, *xi, xi);
  }
  printf("\n");
}

void ivec_map(Ivec *xs, int (*func) (int)) {
  for (size_t i = 0; i < ivec_length (xs); i++) {
    int *xi = ivec_at_internal (xs, i);
    *xi = (*func) (*xi);
  }
}

void ivec_iter(Ivec *xs, void (*func) (int*)) {
  for (size_t i = 0; i < ivec_length (xs); i++) {
    (*func) (ivec_at_internal (xs, i));
  }
}

/* Apply function `fp` to every indexed element of `xs`. */
void ivec_iteri(Ivec *xs, void (*fp) (int, int*)) {
  for (size_t i = 0; i < ivec_length (xs); i++) {
    (*fp) (i, ivec_at_internal (xs, i));
  }
}
